@@ -13,25 +13,29 @@ from NamedAtomicLock import NamedAtomicLock
 from friendly_data.converters import from_df
 from nexinfosys.bin.cli_script import get_valid_name, get_file_url, prepare_base_state, print_issues
 from nexinfosys.command_generators import IType, Issue
+from nexinfosys.command_generators.parser_ast_evaluators import ast_evaluator
+from nexinfosys.command_generators.parser_field_parsers import string_to_ast, arith_boolean_expression
 
 from nexinfosys.common.helper import PartialRetrievalDictionary, any_error_issue, create_dictionary
 from nexinfosys.embedded_nis import NIS
 from nexinfosys.model_services import State
-from nexinfosys.models.musiasem_concepts import Processor, ProcessorsRelationPartOfObservation, Indicator
+from nexinfosys.models.musiasem_concepts import Processor, ProcessorsRelationPartOfObservation, Indicator, \
+    FactorQuantitativeObservation
 from nexinfosys.serialization import deserialize_state
 
 from enbios.common.helper import generate_workbook, list_to_dataframe, get_scenario_name
 from enbios.input import Simulation
 from enbios.input.lci import LCIIndex
 from enbios.input.simulators.sentinel import SentinelSimulation
-from enbios.model import SimStructuralProcessorAttributes
+from enbios.input.simulators.single_csv import AllInACSV
+from enbios.model import SimStructuralProcessorAttributes, g_default_subtech
 from enbios.processing import read_parse_configuration, read_submit_solve_nis_file
 
 
 #####################################################
 # MAIN ENTRY POINT  #################################
 #####################################################
-_idx_cols = ["region", "scenario", "technology", "model", "carrier",
+_idx_cols = ["region", "scenario", "subscenario", "technology", "subtechnology", "model", "carrier",
              "year", "time", "timestep", "unit", "variable", "description"]
 
 
@@ -73,6 +77,7 @@ def parallelizable_process_fragment(param: Tuple[str,  # Fragment label
         """
         nis_idempotent_file = outputs.get("idempotent_nis", None)
         df_indicators = outputs.get("indicators", pd.DataFrame())
+        df_global_indicators = outputs.get("global_indicators", pd.DataFrame())
         df_interfaces = outputs.get("interfaces", pd.DataFrame())
         df_iamc_indicators = outputs.get("iamc_indicators", pd.DataFrame())
         print("Writing results ...")
@@ -102,6 +107,10 @@ def parallelizable_process_fragment(param: Tuple[str,  # Fragment label
                     with open(csv_name, "wt") as f:
                         f.write("Could not obtain indicator values (??)")
 
+        # "system_global_indicators.csv" file (aggregates all fragments)
+        if df_global_indicators is not None:
+            append_fragment_to_file("system_global_indicators.csv", df_global_indicators)
+
         if df_iamc_indicators is not None:
             append_fragment_to_file("iamc_indicators.csv", df_iamc_indicators)
 
@@ -134,7 +143,6 @@ def parallelizable_process_fragment(param: Tuple[str,  # Fragment label
                                keep_fragment_file,
                                generate_nis_fragment_file,
                                generate_interface_results)
-
     end = time.time()  # Stop timing
     print(f"Fragment processed in {end - start} seconds ---------------------------------------")
     write_outputs(outputs)
@@ -159,6 +167,9 @@ class Enviro:
         simulation = None
         if self._cfg["simulation_type"].lower() == "sentinel":
             simulation = SentinelSimulation(self._simulation_files_path)
+        elif self._cfg["simulation_type"].lower() == "single_csv":
+            simulation = AllInACSV(self._simulation_files_path)
+
         # elif self._cfg["simulation_type"].lower() == "calliope":
         #     simulation = CalliopeSimulation(self._simulation_files_path)
         return simulation
@@ -232,6 +243,7 @@ class Enviro:
         if split_by_period and len(times) > 0:
             partition_lists.append([("_d", t) for t in times])
             mandatory_attributes.append("time")
+        # TODO Spores
         if split_by_scenario and len(scenarios) > 0:
             partition_lists.append([("_s", s) for s in scenarios])
             mandatory_attributes.append("scenario")
@@ -291,16 +303,16 @@ class Enviro:
         MAIN entry point of current ENVIRO
         Previously, a Base NIS must have been prepared, see @_prepare_base
 
-        :param n_fragments: If > 0, reduce the number of fragments to the first n_fragments
+        :param n_fragments: number of fragments to process, 0 for "all"
         :param first_fragment: Index of the first fragment to be processed. To obtain an ordered list of fragments, execute first "enbios enviro" with --fragments-list-file option
         :param generate_nis_base_file: True if the Base file should be generated (once) for testing purposes
-        :param generate_nis_fragment_file: True if the current fragment should be dumped into a NIS formatted XLSX file
-        :param generate_interface_results: True if a CSV with values at interfaces should be produced, for each fragment
-        :param generate_indicators: True if a CSV with indicators should be produced, for each fragment
-        :param fragments_list_file: True if a CSV with the list of fragments should be produced
+        :param generate_nis_fragment_file: True to generate a full NIS formatted XLSX file for each fragment
+        :param generate_interface_results: True to generate a CSV with values at interfaces for each fragment
+        :param generate_indicators: True to generate a CSV with indicators for each fragment
+        :param fragments_list_file: True to generate a CSV with the list of fragments
         :param max_lci_interfaces: Max number of LCI interfaces to consider. 0 for all (default 0)
         :param n_cpus: Number of CPUs of the local computer used to perform the process
-        :param just_prepare_base: True to only preparing Base file and exit
+        :param just_prepare_base: True to only prepare (download, parse and execute, then cache; but not solve) Base file and exit
         :param keep_fragment_file: If True, do not remove the minimal NIS file generated and submitted to NIS
         :return:
         """
@@ -321,7 +333,7 @@ class Enviro:
                          f"Base not ready due to errors, exiting. Please check the issues")
             sys.exit(1)
 
-        if just_prepare_base:
+        if just_prepare_base and not fragments_list_file:
             print("Base processed and cached, exiting because 'just_prepare_base == True'")
             sys.exit(1)
 
@@ -337,16 +349,13 @@ class Enviro:
             with open(os.path.join(output_dir, "nis_base.idempotent.xlsx"), "wb") as f:
                 f.write(nis_file)
 
-        # Remove "indicators.csv" and "iamc_indicators if they exist
-        for f_name in ["indicators.csv", "iamc_indicators.csv"]:
-            f_path = os.path.join(output_dir, f_name)
-            if os.path.isfile(f_path):
-                os.remove(f_path)
-
-        # MAIN LOOP - Split simulation in independent fragments, and process them
+        # Read simulation AND Split it in decoupled fragments
         default_time = self._cfg.get("simulation_default_time")
         fragments = sorted([_ for _ in self._read_simulation_fragments(default_time=default_time)],
                            key=operator.itemgetter(0))
+        fragments = [_ for _ in fragments if len(_[3]) > 0]
+
+        # Produce file with an enumeration of fragments
         if fragments_list_file:
             possible_columns = dict(_g="Regions", _d="Periods", _s="Scenarios")
             used_columns = []
@@ -363,13 +372,31 @@ class Enviro:
                 f.write("\n".join(s))
 
         logging.debug(f"Simulation read, and split in {len(fragments)} fragments")
+
+        if just_prepare_base:
+            print("Base processed and cached, exiting because 'just_prepare_base == True'")
+            sys.exit(1)
+
+        # Remove "indicators.csv" and "iamc_indicators if they exist
+        for f_name in ["indicators.csv", "iamc_indicators.csv"]:
+            f_path = os.path.join(output_dir, f_name)
+            if os.path.isfile(f_path):
+                os.remove(f_path)
+
+        # Possibly reduce the list of fragments to process
         if n_fragments > 0 or first_fragment > 0:
             if n_fragments == 0:
                 fragments = fragments[first_fragment:]
             else:
                 fragments = fragments[first_fragment:first_fragment + n_fragments]
             logging.debug(f"{n_fragments} fragment{'s' if n_fragments > 1 else ''} will be processed, starting from fragment {first_fragment}")
-        if n_cpus == 1:  # Serial execution (anyway, function "parallelizable_process_fragment" is valid for both)
+
+        # PROCESS FRAGMENTS - Sequential or embarrassingly parallel
+        # If 0 -> find an appropriate number of CPUs to use. int(80% of # CPUs) if more than 4; 2 for 4; 1 for < 4
+        if n_cpus == 0:
+            n_cpus = int(0.8*cpu_count()) if cpu_count() > 4 else 2 if cpu_count() == 4 else 1
+
+        if n_cpus == 1:
             for i, (frag_label, partial_key, frag_metadata, frag_processors) in enumerate(fragments):
                 # fragment_metadata: dict with regions, years, scenarios in the fragment
                 # fragment_processors: list of processors with their attributes which will be interfaces
@@ -379,12 +406,7 @@ class Enviro:
                                                 development_nis_file,
                                                 generate_nis_fragment_file, generate_interface_results,
                                                 generate_indicators, max_lci_interfaces, keep_fragment_file)
-
-        else:  # Parallel execution
-            # If 0 -> find an appropriate number of CPUs to use
-            if n_cpus == 0:
-                n_cpus = int(0.8*cpu_count()) if cpu_count() > 4 else 2
-
+        else:
             p = Pool(n_cpus)
             p.map(functools.partial(parallelizable_process_fragment,
                                     s_state=serial_state,
@@ -520,6 +542,10 @@ def process_fragment(base_serial_state: bytes,
           if skip_aux_structural AND is_structural(P(Sim)) AND is_not_accounted(P(Sim)) -> P(Sim) = P(P(Sim))
     """
 
+    # Declare (InterfaceType) the original (from simulation) name even if it is renamed (EcoinventCarrierName) and
+    # duplicate the Interface with the original and the changed name
+    write_original_interface = True
+
     def _get_processors_by_type():
         """
         Returns two sets of Processors:
@@ -566,7 +592,7 @@ def process_fragment(base_serial_state: bytes,
         # SimProc(tech, carrier) ->(child-of)-> FunctMuSProc(ParentProc)
         # SimProc(tech, carrier) as StrucMuSProc(tech, carrier2)
 
-    def _find_lci_and_tech_processors(reg: PartialRetrievalDictionary, structural_procs, base_procs, p):
+    def _find_lci_and_tech_processors(reg: PartialRetrievalDictionary, structural_procs, base_procs, p) -> Dict[Processor, List]:
         """
         Find LCI Processor(s) which are assimilated to "p"
           - Processor "p" should exist, as Structural/not-Accounted, with a Parent, [with no Interfaces?]
@@ -596,7 +622,7 @@ def process_fragment(base_serial_state: bytes,
             return lci_proc
 
         # Find structural not accounted matching the sim processor (match tech name and proc carrier name)
-        target = None  # type: Processor
+        targets: Dict[Processor, List] = {}
         tech = p.attrs["technology"].lower()
         carrier_ = p.attrs["carrier"].lower()  # Mandatory
         for proc_name, proc in structural_procs.items():
@@ -609,11 +635,11 @@ def process_fragment(base_serial_state: bytes,
                     elif "EcoinventCarrierName" in proc.attributes:
                         carrier_field = "EcoinventCarrierName"  # It should never enter here
                     if proc.attributes.get(carrier_field, "").lower() == carrier_:
-                        target = proc
-                        break
+                        targets[proc] = []
                 else:
-                    target = proc  # If "tech" matches and "carrier_" is ""
-        if target:
+                    targets[proc] = []  # If "tech" matches and "carrier_" is ""
+
+        for target in targets.keys():
             lci_proc = _find_reference_lci_processor(target, structural_procs)
             if lci_proc is None:
                 # Parent
@@ -621,11 +647,12 @@ def process_fragment(base_serial_state: bytes,
                 for proc_name, proc in base_procs.items():
                     if s1 == proc_name:
                         lci_proc = _find_reference_lci_processor(proc, structural_procs)
+                        targets[target].append(lci_proc)
                         break
-        else:
-            lci_proc = None
+            else:
+                targets[target].append(lci_proc)
 
-        return [lci_proc], target
+        return targets
 
     def _generate_fragment_nis_file(clone_processors, cloners, processors, interfaces, variables, scenarios):
         """
@@ -712,7 +739,7 @@ def process_fragment(base_serial_state: bytes,
                                          f"Country level processor, region '{reg}'", "", "", "", "", reg, reg])
             considered = set()
             for p_name, proc in base_musiasem_procs.items():
-                if len(p_name.split(".")) > 1 and p not in considered:
+                if len(p_name.split(".")) > 1 and p_sim not in considered:
                     considered.add(proc)
                     # InvokingInterface, RequestedInterface, Scale are mandatory; specify something here
                     cloners_list.append((reg, p_name, "Clone", "", "", ""))
@@ -743,7 +770,7 @@ def process_fragment(base_serial_state: bytes,
         else:
             return p, ""
 
-    def _add_processor(proc, matches, processors_list, already_added_processors_set, techs_used_in_regions):
+    def _add_processor(proc, matches, processors_list, already_added_processors_set, techs_used_in_regions, context):
         """
         Inserts elements in "processors_list", directly usable as worksheet rows in a BareProcessors command
 
@@ -752,6 +779,7 @@ def process_fragment(base_serial_state: bytes,
         :param processors_list:
         :param already_added_processors_set:
         :param techs_used_in_regions:
+        :param context:
         :return:
         """
         first_name = None
@@ -780,7 +808,7 @@ def process_fragment(base_serial_state: bytes,
                     already_added_processors_set.add((p_name, parent_name))
         return first_name, parent_name  # p_name if return_simple else first_name, parent_name
 
-    def _add_interfaces(proc, proc_name, proc_parent_name, lci_matches, matching_tech_proc) -> None:
+    def _add_interfaces(proc, proc_name, proc_parent_name, lci_matches, matching_tech_proc, context) -> None:
         """
         Inserts elements in "interfaces", directly usable as worksheet rows in an Interfaces command
 
@@ -789,30 +817,57 @@ def process_fragment(base_serial_state: bytes,
         :param proc_parent_name:
         :param lci_matches: Which LCI processors match
         :param matching_tech_proc: Which technology processor matches
+        :param context: "time", "region", "scenario", ... in which interfaces are being added
         :return:
         """
 
+        def get_tech_simulation_scaling_factor():
+            _ = matching_tech_proc.attributes.get("SimulationScalingFactor", "1.0")
+            ast = string_to_ast(arith_boolean_expression, _)
+            state = State(context)
+            issues = []
+            res, variables = ast_evaluator(ast, state, None, issues)
+            if res:
+                return res
+            else:
+                raise ValueError(f"Could not evaluate {_}. Variables: {variables}. Issues: {issues}")
+
         relative_to = None
         tech_output_name = matching_tech_proc.attributes.get("SimulationVariable")
         tech_desired_output_name = matching_tech_proc.attributes.get("EcoinventCarrierName")
+        tech_simulation_scaling_factor = get_tech_simulation_scaling_factor()
         tech_output_to_spold_factor = float(matching_tech_proc.attributes.get("SimulationToEcoinventFactor", "1.0"))
         observer = get_scenario_name("o", scenario)
+        # Interfaces from MuSIASEM tech
+        for i in matching_tech_proc.factors:
+            for o in i.observations:
+                if isinstance(o, FactorQuantitativeObservation):
+                    interfaces.append([name, i.name, "", "", "", i.orientation, "", "", "", "", o.value,
+                                       "", "", "", "", "", "", o.attributes["time"],
+                                       o.observer.name if o.observer else "", "", ""])
+
         # Interfaces from simulation
         for i in set(proc.attrs.keys()).difference(_idx_cols):
             i_name = i
+            input_i_name = i_name
             v = proc.attrs[i]
             if is_main_flow(i, tech_output_name):
                 if tech_desired_output_name:
                     i_name = tech_desired_output_name
-                v *= tech_output_to_spold_factor  # Change scale
+                v *= tech_output_to_spold_factor * tech_simulation_scaling_factor  # Change scale
                 relative_to = i_name  # All LCI interfaces will be relative to main output
                 orientation = "Output"
             else:
                 orientation = get_flow_orientation(i)
             interfaces.append([proc_name, i_name, "", "", "", orientation, "", "", "", "", v,
                                "", "", "", "", "", "", time_, observer, "", ""])
+            if write_original_interface and input_i_name != i_name:
+                interfaces.append([proc_name, input_i_name, "", "", "", orientation, "", "", "", "", v,
+                                   "", "", "", "", "", "", time_, observer, "", ""])
             if i_name not in variables:
                 variables[i_name] = dict(main_flow=is_main_flow(i), orientation=orientation, unit=get_flow_unit(i))
+            if write_original_interface and input_i_name not in variables:
+                variables[input_i_name] = dict(main_flow=False, orientation=orientation, unit=get_flow_unit(i))
 
         if lci_matches[0] is None or len(lci_matches) == 0:
             print(f"{proc.attrs} does not have an LCI processor associated. "
@@ -963,47 +1018,61 @@ def process_fragment(base_serial_state: bytes,
         # Modify all parent processors system to region "r"
         if len(fragment_metadata["regions"]) == 1:
             r = next(iter(fragment_metadata["regions"]))
-            for p in dendrogram_musiasem_procs.values():
-                p.processor_system = r
+            for p_sim in dendrogram_musiasem_procs.values():
+                p_sim.processor_system = r
 
     # FOR EACH SIMULATION PROCESSOR (of the fragment)
     iamc_codes_techs = create_dictionary()
-    for p in fragment_processors:
+    for p_sim in fragment_processors:
         # Update time and scenario
         # (they can change from entry to entry, but for MuSIASEM the same functional Processor is used)
-        time_ = p.attrs.get("time", p.attrs.get("year", default_time))
-        scenario = p.attrs.get("scenario", "")
-        region = p.attrs.get("region", "")
-        carrier = p.attrs.get("carrier", "")
-        tech = p.attrs.get("technology", "")
+        time_ = p_sim.attrs.get("time", p_sim.attrs.get("year", default_time))
+        scenario = p_sim.attrs.get("scenario", "")
+        region = p_sim.attrs.get("region", "")
+        carrier = p_sim.attrs.get("carrier", "")
+        tech = p_sim.attrs.get("technology", "")
+        subtech = p_sim.attrs.get("subtechnology", g_default_subtech)
+        subscenario = p_sim.attrs.get("subscenario", "")
+        context = dict(time=time_, scenario=scenario, region=region, carrier=carrier, tech=tech, subtech=subtech)
         if carrier == "":
-            print(f"Processor {p.attrs} ignored because carrier is not defined")
+            print(f"'carrier' is not defined for processor {p_sim.attrs}, ignored")
             continue  # Ignore processors not having carrier defined
+        from enbios import subtech_supported
+        if subtech_supported and subtech != g_default_subtech:
+            # TODO Add subtech support
+            print(f"'subtechnology' field is not really supported, processor {p_sim.attrs} ignored")
+            continue  # Ignore processors having subtechnology defined
 
         # Find "LCI" and "Reference Tech" matching Processor(s)
         # (considering "carrier" - "EcoinventCarrierName" if it is defined)
-        matching_lci, matching_tech = _find_lci_and_tech_processors(prd,
-                                                                    structural_not_accounted_procs,
-                                                                    dendrogram_musiasem_procs,
-                                                                    p)
-        if matching_tech is None:
-            print(f"Could not find a tech processor matching {p.attrs}. Skipping")
+        # matching_lci, matching_not_accounted_reference_tech = _find_lci_and_tech_processors(prd,
+        #                                                                                     structural_not_accounted_procs,
+        #                                                                                     dendrogram_musiasem_procs,
+        #                                                                                     p_sim)
+        matching_not_accounted_reference_techs = _find_lci_and_tech_processors(prd,
+                                                                               structural_not_accounted_procs,
+                                                                               dendrogram_musiasem_procs,
+                                                                               p_sim)
+        if len(matching_not_accounted_reference_techs) == 0:
+            print(f"Could not find a tech processor matching {p_sim.attrs}. Skipping")
             continue
 
-        # Parent processor(s) in "MUSIASEM Dendrogram"
-        musiasem_matches = _find_parents(prd, structural_not_accounted_procs, matching_tech)
-        if musiasem_matches is None or len(musiasem_matches) == 0:
-            print(f"{p} has no parents. Cannot account it, skipped.")
-            continue
+        for matching_not_accounted_reference_tech, matching_lci in matching_not_accounted_reference_techs.items():
+            # Parent processor(s) in "MUSIASEM Dendrogram"
+            musiasem_matches = _find_parents(prd, structural_not_accounted_procs, matching_not_accounted_reference_tech)
+            if musiasem_matches is None or len(musiasem_matches) == 0:
+                print(f"{p_sim} has no parents. Cannot account it, skipped.")
+                continue
 
-        iamc_codes_techs[tech] = matching_tech.attributes.get("iamccode", matching_tech.full_hierarchy_name.replace(".", "|"))
+            iamc_codes_techs[tech] = matching_not_accounted_reference_tech.attributes.get(
+                "iamccode", matching_not_accounted_reference_tech.full_hierarchy_name.replace(".", "|"))
 
-        # Add PROCESSOR(S)
-        name, parent = _add_processor(p, musiasem_matches, processors, already_added_processors, techs_used_in_regions)
+            # Add PROCESSOR(S)
+            name, parent = _add_processor(p_sim, musiasem_matches, processors, already_added_processors, techs_used_in_regions, context)
 
-        # Add INTERFACES (FROM Simulation AND FROM LCI), if we had a processor added (a name)
-        if name:  # At least a match? -> add Interfaces, combining
-            _add_interfaces(p, name, parent, matching_lci, matching_tech)
+            # Add INTERFACES (FROM Simulation AND FROM LCI), if we had a processor added (a name)
+            if name:  # At least a match? -> add Interfaces, combining
+                _add_interfaces(p_sim, name, parent, matching_lci, matching_not_accounted_reference_tech, context)
 
     print(f"Generating NIS file ...")  # for: {fragment_metadata}")
 
@@ -1022,6 +1091,7 @@ def process_fragment(base_serial_state: bytes,
                 _.append("flow_graph_solution")
             if generate_nis_fragment_file:
                 _.append("model")
+            _.append("flow_graph_global_indicators")
 
             issues, new_state, ds = run_nis_for_results(nis_file_name, development_nis_file, state, _)
             if ds:
@@ -1033,6 +1103,7 @@ def process_fragment(base_serial_state: bytes,
                     outputs["interfaces"] = ds[1]
                 if generate_nis_fragment_file:
                     outputs["idempotent_nis"] = ds[2] if len(ds) == 3 else ds[1]
+                outputs["global_indicators"] = ds[-1]
             else:
                 print(f"There were issues processing fragment file: {nis_file_name}")
                 print_issues("Solving fragment", nis_file_name, issues, f"Please check the issues resulting from solving fragment '{nis_file_name}'")
